unit UnitRobot;

interface

uses Windows, KOL;

const
    MaxSz = 40;
var
    MaxX: Integer = 40;
    MaxY: Integer = 40;

type
  TDirection = ( d_Left, d_Top, d_Right, d_Bottom );
const
  BackDir: array[ TDirection ] of TDirection = ( d_Right, d_Bottom, d_Left, d_Top );
  LeftDir: array[ TDirection ] of TDirection = ( d_Bottom, d_Left, d_Top, d_Right );
  RightDir: array[ TDirection ] of TDirection = ( d_Top, d_Right, d_Bottom, d_Left );
type
  PRobot = ^TRobot;
  TRobot = object( TObj )
  Army: Integer; // 1..4
  Direction: TDirection;
  Energy: Integer; // 1..100
  Vars: array[ 'X'..'Z' ] of Integer;
  LastMsgCoords: TPoint; // X in 1..MaxX, Y in 1..MaxY
  LastMsgLabel: String;
  HeardCoord: TPoint;
  HeardLabel: String;
  InstructionPt: TPoint; // Y =    (0..), X =    (0..)
  CallStack: PList;

  ProgramText: PStrList;
  GoForward: Boolean;
  DoAttack: Boolean;
  Ready: Boolean;

  EnergyAdd: Integer;

  procedure Init; virtual;
  destructor Destroy; virtual;
  function Distance: Integer;
  procedure Check( Value: Integer; S: PChar );
  procedure WorkWithVar( Value: PInteger; S: PChar );
  procedure SendMessage( S: PChar );

  procedure TurnToHeard;
  function CurrentCommand: PChar;
  procedure SkipToSpc;
  procedure GotoLabel( const GoLabel: String = '' );
  function PositionOnDesk: TPoint;
  function EnemyForward: Boolean;
  function EnemyCount: Integer;
  function AllyCount: Integer;
  function AnyForward: Boolean;
  function Attack: Boolean;
  function Transfer( S: PChar ): Boolean;
  end;

var Desk: array[ 1..MaxSz, 1..MaxSz ] of PRobot;

function PositionForward( const P: TPoint; Direction: TDirection ): TPoint;
function GetNum( var S: PChar ): Integer;

implementation

const FwIncPos: array[ TDirection ] of TPoint = ( ( X:-1; Y:0 ), ( X:0; Y:-1 ),
      (X:1; Y:0), (X:0; Y:1) );

function PositionForward( const P: TPoint; Direction: TDirection ): TPoint;
begin
  Result.X := P.X + FwIncPos[ Direction ].X;
  if Result.X > MaxX then Result.X := 1 else
  if Result.X < 1    then Result.X := MaxX;
  Result.Y := P.Y + FwIncPos[ Direction ].Y;
  if Result.Y > MaxY then Result.Y := 1 else
  if Result.Y < 1    then Result.Y := MaxY;
end;

function GetNum( var S: PChar ): Integer;
begin
  Result := 0;
  while S^ in [ '0'..'9' ] do
  begin
    Result := Result * 10 + Ord( S^ ) - Ord( '0' );
    Inc( S );
  end;
end;

{ TRobot }

function TRobot.AllyCount: Integer;
var P, P1: TPoint;
    R2: PRobot;
begin
  Result := 0;
  P1 := PositionOnDesk;
  P := PositionForward( P1, d_Left );
  R2 := Desk[ P.X, P.Y ];
  if (R2 <> nil) and (R2.Army = Army) then Inc( Result );
  P := PositionForward( P1, d_Top );
  R2 := Desk[ P.X, P.Y ];
  if (R2 <> nil) and (R2.Army = Army) then Inc( Result );
  P := PositionForward( P1, d_Right );
  R2 := Desk[ P.X, P.Y ];
  if (R2 <> nil) and (R2.Army = Army) then Inc( Result );
  P := PositionForward( P1, d_Bottom );
  R2 := Desk[ P.X, P.Y ];
  if (R2 <> nil) and (R2.Army = Army) then Inc( Result );
end;

function TRobot.AnyForward: Boolean;
var P: TPoint;
    R2: PRobot;
begin
  P := PositionOnDesk;
  P := PositionForward( P, Direction );
  R2 := Desk[ P.X, P.Y ];
  Result := R2 <> nil;
end;

function TRobot.Attack: Boolean;
var P: TPoint;
    R2: PRobot;
    E: Integer;
begin
  Result := FALSE;
  P := PositionOnDesk;
  P := PositionForward( P, Direction );
  R2 := Desk[ P.X, P.Y ];
  if (R2 = nil) or (R2.Army = Army) then
  begin
    //Dec( Energy, 0 );
    if R2 <> nil then
    begin
      if Direction = R2.Direction then E := 20 //   
      else if Direction = BackDir[ R2.Direction ] then E := 5 //   
      else E := 10;
      Inc( Energy, E );
      Dec( R2.Energy, E );
    end;
    Dec( Energy, 2 );
  end
    else
  begin
      if Direction = R2.Direction then E := 20 //   
      else if Direction = BackDir[ R2.Direction ] then E := 5 //   
      else E := 10;
      Inc( Energy, E );
      Dec( R2.Energy, E );
      Result := TRUE;
  end;
end;

procedure TRobot.Check(Value: Integer; S: PChar);
var C: Char;
    Flag: Boolean;
    N: Integer;
begin
 while (S^ <> #0) and (S^ in [ 'A'..'Z' ]) do Inc( S );
 if S^ = '.' then Inc( S );
 Flag := FALSE;
 if S^ in [ '<', '>' ] then
 begin
   C := S^;
   Inc( S );
   if S^ = '.' then Inc( S );
   if S^ in [ '0'..'9' ] then
   begin
     N := GetNum( S );
     CASE C OF
     '<': Flag := Value < N;
     '>': Flag := Value > N;
     END;
   end
     else
   if S^ in [ 'X'..'Z' ] then
   begin
     CASE C OF
     '<': Flag := Value < Vars[ S^ ];
     '>': Flag := Value > Vars[ S^ ];
     END;
   end;
 end
   else
 if S^ in [ '0'..'9' ] then
 begin
   N := GetNum( S );
   Flag := Value < N;
 end
   else
 if S^ in [ 'X'..'Z' ] then
 begin
   Flag := Value < Vars[ S^ ];
 end;
 if Flag then GotoLabel
         else SkipToSpc;
end;

function TRobot.CurrentCommand: PChar;
var S: PChar;
begin
  if @ self = nil then
  asm
    int 3
  end;
  S := nil;
  while InstructionPt.Y < ProgramText.Count do
  begin
    S := ProgramText.ItemPtrs[ InstructionPt.Y ] + InstructionPt.X;
    if InstructionPt.X = 0 then
    begin
      if (S^ <> #0) and (S^ in [ 'a'..'z', ''..'', '_' ]) then
      begin
        while (S^ <> #0) and (S^ <> ':') do
        begin
          Inc( InstructionPt.X );
          Inc( S );
        end;
      end;
    end;
    while S^ <> #0 do
    begin
      if S^ = '*' then break;
      if not (S^ in [ 'A'..'Z' ]) then begin Inc( S ); Inc( InstructionPt.X ); end
                                else break;
    end;
    if S^ in [ '*', #0 ] then
    begin
      Inc( InstructionPt.Y );
      InstructionPt.X := 0;
    end
    else
      break;
  end;
  Result := S;
end;

destructor TRobot.Destroy;
begin
  CallStack.Free;
  LastMsgLabel := '';
  HeardLabel := '';
  inherited;
end;

function TRobot.Distance: Integer;
var P: TPoint;
    I: Integer;
begin
  Result := 0;
  if LastMsgCoords.X = 0 then Exit;
  P := PositionOnDesk;
  Result := Abs( P.X - LastMsgCoords.X );
  if Result > MaxX div 2 then
    Result := MaxX - Result;
  I := Abs( P.Y - LastMsgCoords.Y );
  if I > MaxY div 2 then
    I := MaxY - I;
  if I > Result then Result := I;
end;

function TRobot.EnemyCount: Integer;
var P, P1: TPoint;
    R2: PRobot;
begin
  Result := 0;
  P1 := PositionOnDesk;
  P := PositionForward( P1, d_Left );
  R2 := Desk[ P.X, P.Y ];
  if (R2 <> nil) and (R2.Army <> Army) then Inc( Result );
  P := PositionForward( P1, d_Top );
  R2 := Desk[ P.X, P.Y ];
  if (R2 <> nil) and (R2.Army <> Army) then Inc( Result );
  P := PositionForward( P1, d_Right );
  R2 := Desk[ P.X, P.Y ];
  if (R2 <> nil) and (R2.Army <> Army) then Inc( Result );
  P := PositionForward( P1, d_Bottom );
  R2 := Desk[ P.X, P.Y ];
  if (R2 <> nil) and (R2.Army <> Army) then Inc( Result );
end;

function TRobot.EnemyForward: Boolean;
var P: TPoint;
    R2: PRobot;
begin
  P := PositionOnDesk;
  P := PositionForward( P, Direction );
  R2 := Desk[ P.X, P.Y ];
  Result := (R2 <> nil) and (R2.Army <> Army);
end;

procedure TRobot.GotoLabel( const GoLabel: String = '' );
var S: PChar;
    L: String;
    I: Integer;
begin
  if GoLabel <> '' then L := GoLabel
  else
  begin
    S := CurrentCommand;
    while (S^ <> #0) and (S^ in ['A'..'Z']) do Inc( S );
    if S^ in [ '.' ] then Inc( S );
    if S^ in [ '<', '>', '+', '-', '=' ] then Inc( S );
    if S^ in [ '.' ] then Inc( S );
    while (S^ <> #0) and (S^ in ['A'..'Z','0'..'9']) do Inc( S );
    if S^ in [ '.' ] then Inc( S );
    L := '';
    if S^ in [ 'a'..'z', ''..'', '_' ] then
    begin
      L := '';
      while (S^ <> #0) and (S^ > ' ') do
      begin
        L := L + S^;
        Inc( S );
      end;
    end;
  end;
  if L <> '' then
  begin
    L := L + ':';
    for I := 0 to ProgramText.Count-1 do
    begin
      S := ProgramText.ItemPtrs[ I ];
      if (S^ = L[ 1 ]) and StrIsStartingFrom( S, PChar( L ) ) then
      begin
        InstructionPt.Y := I;
        InstructionPt.X := Length( L );
        Exit;
      end;
    end;
  end;
  SkipToSpc;
end;

procedure TRobot.Init;
begin
  inherited;
  CallStack := NewList;
end;

function TRobot.PositionOnDesk: TPoint;
var I, J: Integer;
begin
  for I := 1 to MaxX do
  for J := 1 to MaxY do
    if Desk[ I, J ] = @ Self then
    begin
      Result := MakePoint( I, J );
      Exit;
    end;
  Result := MakePoint( 0, 0 );
end;

procedure TRobot.SendMessage(S: PChar);
var N: Integer;
    L: String;
    I, J, X, Y: Integer;
    P: TPoint;
    R2: PRobot;
begin
  while (S^ <> #0) and (S^ in ['A'..'Z']) do Inc( S );
  if S^ in [ '.' ] then Inc( S );
  N := 4;
  if S^ in [ '0'..'9' ] then
    N := GetNum( S );
  if S^ in [ '.' ] then Inc( S );
  L := '';
  if S^ in [ 'a'..'z', ''..'', '_' ] then
  begin
    while (S^ <> #0) and (S^ > ' ') do
    begin
      L := L + S^;
      Inc( S );
    end;
  end;
  P := PositionOnDesk;
  for I := P.Y-N to P.Y+N do
  for J := P.X-N to P.X+N do
  begin
    X := J;
    if X < 1 then X := X + MaxX else
    if X > MaxX then X := X - MaxX;
    Y := I;
    if Y < 1 then Y := Y + MaxY else
    if Y > MaxY then Y := Y - MaxY;
    R2 := Desk[ X, Y ];
    if R2 = nil then Continue;
    if R2 = @ Self then Continue;
    if R2.Army = Army then
    begin
      R2.LastMsgCoords := P;
      R2.LastMsgLabel := L;
    end;
  end;
  SkipToSpc;
end;

procedure TRobot.SkipToSpc;
var S: PChar;
    C: Char;
begin
  if InstructionPt.Y >= ProgramText.Count then
    InstructionPt := MakePoint( 0, 0 );
  S := ProgramText.ItemPtrs[ InstructionPt.Y ] + InstructionPt.X;
  while (S^ <> #0) and (S^ > ' ') do
  begin
    C := S^;
    Inc( S );
    Inc( InstructionPt.X );
    if (C = ':') and (S^ > ' ') and (S^ <> '*') then Exit;
    while (S^ in [#0, '*']) do
    begin
      Inc( InstructionPt.Y );
      InstructionPt.X := 0;
      if InstructionPt.Y >= ProgramText.Count then Exit;
      S := ProgramText.ItemPtrs[ InstructionPt.Y ] + InstructionPt.X;
      if not (S^ in [ 'a'..'z', ''..'', '_', #0, '*' ]) then Exit;
    end;
  end;
end;

function TRobot.Transfer( S: PChar ): Boolean;
var P: TPoint;
    R2: PRobot;
    E: Integer;
begin
  Result := FALSE;
  P := PositionOnDesk;
  P := PositionForward( P, Direction );
  R2 := Desk[ P.X, P.Y ];
  E := 0;
  while S^ in [ 'A'..'Z' ] do Inc( S );
  if S^ = '.' then Inc( S );
  if S^ in [ '0'..'9' ] then
    E := GetNum( S );
  if (R2 = nil) or (R2.Army <> Army) then
  begin
    Dec( Energy, Max( 1, E div 10 ) );
  end
    else
  begin
      Dec( Energy, E );
      Inc( R2.Energy, E );
      Result := TRUE;
  end;
end;

procedure TRobot.TurnToHeard;
var P: TPoint;
    dX, dY: Integer;
begin
  if HeardLabel = '' then Exit;
  P := PositionOnDesk;
  dX := HeardCoord.X - P.X;
  dY := HeardCoord.Y - P.Y;
  if dX > MaxX div 2 then
    dX := dX - MaxX
  else
  if dX < - MaxX div 2 then
    dX := dX + MaxX;
  if dY > MaxY div 2 then
    dY := dY - MaxY
  else
    dY := dY + MaxY;
  if Abs( dY ) > Abs( dX ) then
  begin
    if dY > 0 then Direction := d_Bottom else
    if dY < 0 then Direction := d_Top;
  end
    else
  begin
    if dX < 0 then Direction := d_Left else
    if dX > 0 then Direction := d_Right;
  end;
end;

procedure TRobot.WorkWithVar(Value: PInteger; S: PChar);
var C: Char;
    Flag: Boolean;
    N: Integer;
begin
 while (S^ <> #0) and (S^ in [ 'A'..'Z' ]) do Inc( S );
 if S^ = '.' then Inc( S );
 Flag := FALSE;
 if S^ in [ '=', '+', '-' ] then
 begin
   C := S^;
   Inc( S );
   if S^ = '.' then Inc( S );
   N := 0;
   if S^ in [ '0'..'9' ] then
     N := GetNum( S )
   else if S^ in [ 'D', 'E', 'M', 'N', 'X', 'Y', 'Z' ] then
   begin
     CASE S^ OF
     'D': N := Distance;
     'E': N := Energy;
     'M': N := AllyCount;
     'N': N := EnemyCount;
     'X', 'Y', 'Z': N := Vars[ S^ ];
     END;
   end;
   while (S^ <> #0) and (S^ in [ 'A'..'Z' ]) do Inc( S );
   CASE C OF
   '=': Value^ := N;
   '+': Value^ := Value^ + N;
   '-': Value^ := Value^ - N;
   END;
   Flag := Value^ < 0;
 end
   else
 if S^ in [ '<', '>' ] then
 begin
   C := S^;
   Inc( S );
   if S^ = '.' then Inc( S );
   if S^ in [ '0'..'9' ] then
   begin
     N := GetNum( S );
     CASE C OF
     '<': Flag := Value^ < N;
     '>': Flag := Value^ > N;
     END;
   end
     else
   if S^ in [ 'X'..'Z' ] then
   begin
     CASE C OF
     '<': Flag := Value^ < Vars[ S^ ];
     '>': Flag := Value^ > Vars[ S^ ];
     END;
   end;
 end
   else
 if S^ in [ '0'..'9' ] then
 begin
   N := GetNum( S );
   Flag := Value^ < N;
 end
   else
 if S^ in [ 'X'..'Z' ] then
 begin
   Flag := Value^ < Vars[ S^ ];
 end;
 if Flag then GotoLabel
         else SkipToSpc;
end;

end.
